"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const uuid = require("uuid");
class ProcessInstance {
    constructor(processKey, messageBusService, processable) {
        this._messageBusService = undefined;
        this._processKey = undefined;
        this._processable = undefined;
        this._participantSubscription = undefined;
        this._nextTaskDef = undefined;
        this._nextTaskEntity = undefined;
        this._taskChannelName = undefined;
        this._tokenData = undefined;
        this._participantId = undefined;
        this._eventSubscription = undefined;
        this._eventChannelName = undefined;
        this._messageBusService = messageBusService;
        this._processKey = processKey;
        this._processable = processable;
        this._participantId = uuid.v4();
    }
    get messageBusService() {
        return this._messageBusService;
    }
    get processable() {
        return this._processable;
    }
    get processKey() {
        return this._processKey;
    }
    get nextTaskDef() {
        return this._nextTaskDef;
    }
    set nextTaskDef(nextTaskDef) {
        this._nextTaskDef = nextTaskDef;
    }
    get nextTaskEntity() {
        return this._nextTaskEntity;
    }
    set nextTaskEntity(nextTaskEntity) {
        this._nextTaskEntity = nextTaskEntity;
    }
    get taskChannelName() {
        return this._taskChannelName;
    }
    set taskChannelName(taskChannelName) {
        this._taskChannelName = taskChannelName;
    }
    get eventChannelName() {
        return this._eventChannelName;
    }
    get participantId() {
        return this._participantId;
    }
    get tokenData() {
        return this._tokenData;
    }
    async start(context, token) {
        const msg = this.messageBusService.createDataMessage({
            action: 'start',
            key: this.processKey,
            token,
        }, context, {
            participantId: this.participantId,
        });
        await this._listenToParticipantChannel();
        await this.messageBusService.publish('/processengine', msg);
        return this;
    }
    async continueProcess(userTaskEntity, context) {
        const meta = {
            isRef: true,
            type: 'UserTask',
            namespace: undefined,
            isNew: false,
        };
        const data = {
            id: userTaskEntity.id,
            _meta: meta,
        };
        const msg = this.messageBusService.createDataMessage({
            action: 'continue',
            nodeInstanceRef: data,
        }, context, {
            participantId: this.participantId,
        });
        await this._listenToParticipantChannel();
        await this.messageBusService.publish('/processengine', msg);
        return this;
    }
    async stop() {
        this.nextTaskDef = null;
        this.nextTaskEntity = null;
        this.taskChannelName = null;
        this._tokenData = null;
        await this._participantSubscription.cancel();
        return;
    }
    async restart(context, token) {
        await this.stop();
        await this.start(context, token);
        return;
    }
    async doCancel(context) {
        const msg = this.messageBusService.createDataMessage({
            action: 'event',
            eventType: 'cancel',
            data: this._tokenData,
        }, context, {
            participantId: this.participantId,
        });
        await this.messageBusService.publish(this.taskChannelName, msg);
        return;
    }
    async doEvent(context, eventData) {
        const msg = this.messageBusService.createDataMessage({
            action: 'event',
            eventType: 'data',
            data: eventData,
        }, context, {
            participantId: this.participantId,
        });
        await this.messageBusService.publish(this.taskChannelName, msg);
        return;
    }
    async doError(context, error) {
        const msg = this.messageBusService.createDataMessage({
            action: 'event',
            eventType: 'error',
            data: error,
        }, context, {
            participantId: this.participantId,
        });
        await this.messageBusService.publish(this.taskChannelName, msg);
        return;
    }
    async doProceed(context) {
        const msg = this.messageBusService.createDataMessage({
            action: 'proceed',
            token: this._tokenData,
        }, context, {
            participantId: this.participantId,
        });
        await this.messageBusService.publish(this.taskChannelName, msg);
        return;
    }
    async _listenToParticipantChannel() {
        const participantChannelName = '/participant/' + this.participantId;
        this._participantSubscription = await this.messageBusService.subscribe(participantChannelName, async (message) => {
            if (!this.processable) {
                throw new Error('no processable defined to handle activities!');
            }
            else if (message && message.data && message.data.action) {
                await this._evaluateAction(message.data.action, message.data.data);
            }
        });
    }
    async _evaluateAction(action, messageData) {
        if (action) {
            switch (action) {
                case 'userTask':
                    await this._setNewTask(messageData);
                    const uiName = messageData.uiName;
                    const uiConfig = messageData.uiConfig;
                    this._tokenData = messageData.uiData || {};
                    await this.processable.handleUserTask(this, uiName, uiConfig, this._tokenData);
                    break;
                case 'manualTask':
                    await this._setNewTask(messageData);
                    const taskName = messageData.uiName;
                    const taskConfig = messageData.uiConfig;
                    this._tokenData = messageData.uiData || {};
                    await this.processable.handleManualTask(this, taskName, taskConfig, this._tokenData);
                    break;
                case 'endEvent':
                    this._tokenData = messageData || {};
                    await this.processable.handleEndEvent(this, this._tokenData);
                    await this.stop();
                    break;
            }
        }
    }
    async _setNewTask(taskMessageData) {
        this.nextTaskDef = taskMessageData.userTaskEntity.nodeDef;
        this.nextTaskEntity = taskMessageData.userTaskEntity;
        this.taskChannelName = '/processengine/node/' + this.nextTaskEntity.id;
        this._eventChannelName = '/processengine_api/event/' + this.nextTaskEntity.id;
        this._eventSubscription = await this.messageBusService.subscribe(this.eventChannelName, async (message) => {
            switch (message.data.action) {
                case 'event':
                    const eventType = message.data.eventType;
                    const eventData = message.data.data || {};
                    switch (eventType) {
                        case 'cancel':
                            await this.processable.handleCancel(this);
                            break;
                        default:
                            this._tokenData.current = eventData;
                            await this.processable.handleEvent(this, eventType, eventData);
                            break;
                    }
            }
        });
    }
}
exports.ProcessInstance = ProcessInstance;

//# sourceMappingURL=process_instance.js.map
